from ckanext.drupal_idp.utils import Details
import mock

import pytest

import ckanext.drupal_idp.utils as utils


@pytest.fixture
def details_data():
    return {
        "name": "test",
        "email": "test@example.net",
        "id": 123,
        "roles": [],
    }


@pytest.fixture
def with_no_drupal_session(monkeypatch):
    monkeypatch.setattr(
        utils, "get_user_details", mock.Mock(return_value=None)
    )


@pytest.fixture
def with_drupal_session(monkeypatch, details_data):
    func = mock.Mock(return_value=Details(**details_data))
    monkeypatch.setattr(utils, "get_user_details", func)



import logging

import ckan.plugins as plugins
import ckan.plugins.toolkit as tk

from ckanext.drupal_idp.logic import action
from ckanext.drupal_idp.logic import auth
from ckanext.drupal_idp import helpers, utils, drupal, cli



log = logging.getLogger(__name__)


class DrupalIdpPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IAuthenticator, inherit=True)
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IActions)
    plugins.implements(plugins.IAuthFunctions)
    plugins.implements(plugins.IClick)

    # IClick

    def get_commands(self):
        return cli.get_commands()

    # IActions

    def get_actions(self):
        return action.get_actions()

    # IAuthFunctions

    def get_auth_functions(self):
        return auth.get_auth_functions()

    # ITemplateHelpers

    def get_helpers(self):
        return helpers.get_helpers()

    # IAuthenticator

    def identify(self):
        """This does drupal authorization.
        The drupal session contains the drupal id of the logged in user.
        We need to convert this to represent the ckan user."""
        cookie_sid = tk.request.cookies.get(utils.session_cookie_name())
        if not cookie_sid:
            return
        sid = utils.decode_sid(cookie_sid)
        details = utils.get_user_details(sid)
        if not details:
            return
        try:
            user = utils.get_or_create_from_details(details)
            if utils.is_synchronization_enabled():
                user = utils.synchronize(user, details)
        except tk.ValidationError as e:
            log.error(
                f"Cannot create user {details.name}<{details.email}>:"
                f" {e.error_summary}"
            )
            return
        tk.c.user = user["name"]

    # IConfigurer

    def update_config(self, config_):
        # If DB config is missing, the following line will raise
        # CkaneConfigurationException and won't allow server to start
        drupal.db_url()



from typing import Optional

import ckan.model as model

from .utils import DrupalId


def get_helpers():
    return {
        "drupal_idp_drupal_id": drupal_id,
    }


def drupal_id(id_or_name: str) -> Optional[DrupalId]:
    user = (
        model.Session.query(
            model.User.plugin_extras["drupal_idp"]["id"].label("drupal_id")
        )
        .filter(
            (model.User.id == id_or_name) | (model.User.name == id_or_name)
        )
        .one_or_none()
    )
    if user:
        return user.drupal_id



import pytest

import ckan.plugins.toolkit as tk
from ckan.tests.helpers import call_action

import ckanext.drupal_idp.utils as utils


class TestUserShow:
    def test_id_is_mandatory(self):
        with pytest.raises(tk.ValidationError):
            call_action('drupal_idp_user_show')

    def test_no_user_raises_an_error(self):
        with pytest.raises(tk.ObjectNotFound):
            call_action('drupal_idp_user_show', id=10)

    @pytest.mark.usefixtures("clean_db")
    def test_no_user_raises_an_error(self, details_data):
        details = utils.Details(**details_data)
        userdict = utils.get_or_create_from_details(details)

        user = call_action('drupal_idp_user_show', id=details.id)
        assert user["id"] == userdict["id"]



import click

import ckan.model as model
import ckan.plugins.toolkit as tk

import ckanext.drupal_idp.utils as utils

def get_commands():
    return [drupal_idp]


@click.group(short_help='ckanext-drupal-idp CLI')
def drupal_idp():
    """
    """
    pass

@drupal_idp.group()
def user():
    """User management
    """
    pass

@user.command()
def list():
    """List all users with DrupalID
    """
    users = model.Session.query(model.User).filter(model.User.plugin_extras.has_key('drupal_idp'))
    for user in users:
        click.echo(f"{user.name}:")
        extras = user.plugin_extras['drupal_idp']
        click.echo(f"\tID: {extras['id']}")
        click.echo(f"\tRoles: {', '.join(extras['roles'])}")
        click.echo(f"\tEmail: {extras['email']}")
        click.echo()



import pytest

from ckan.exceptions import CkanConfigurationException

import ckanext.drupal_idp.utils as utils
import ckanext.drupal_idp.drupal as drupal

class TestDbUrl:
    def test_default_value(self):
        assert drupal.db_url()

    @pytest.mark.ckan_config(utils.CONFIG_DB_URL, "")
    def test_exception_with_missing_config(self):
        with pytest.raises(CkanConfigurationException):
            drupal.db_url()



import logging
import abc
from typing import Any, Dict, Iterable, List, Optional

import sqlalchemy as sa
from sqlalchemy.exc import OperationalError

import ckan.plugins.toolkit as tk
from ckan.exceptions import CkanConfigurationException

import ckanext.drupal_idp.utils as utils

log = logging.getLogger(__name__)


def db_url() -> str:
    url = tk.config.get(utils.CONFIG_DB_URL)
    if not url:
        raise CkanConfigurationException(
            f"drupal_idp plugin requires {utils.CONFIG_DB_URL} config option."
        )
    return url


class BaseDrupal(metaclass=abc.ABCMeta):
    def __init__(self, url: str):
        self.engine = sa.create_engine(url)

    @abc.abstractmethod
    def get_user_by_sid(self, sid: str) -> Optional[Any]:
        ...

    @abc.abstractmethod
    def get_user_roles(self, uid: utils.DrupalId) -> List[str]:
        ...


class Drupal9(BaseDrupal):
    def get_user_by_sid(self, sid: str) -> Optional[Any]:
        try:
            user = self.engine.execute(
                """
            SELECT d.name name, d.mail email, d.uid id
            FROM sessions s
            JOIN users_field_data d
            ON s.uid = d.uid
            WHERE s.sid = %s
            """,
                [sid],
            ).first()
        except OperationalError:
            log.exception("Cannot get a user from Drupal's database")
            return
        # check if session has username,
        # otherwise is unauthenticated user session
        if user and user.name:
            return user

    def get_user_roles(self, uid: utils.DrupalId) -> List[str]:
        query = self.engine.execute(
            """
                 SELECT roles_target_id name
                 FROM user__roles
                 WHERE bundle = 'user' AND entity_id = %s
                 """,
            [uid],
        )
        return [role.name for role in query]


_mapping = {"9": Drupal9}


def get_adapter(version: str) -> BaseDrupal:

    return _mapping[version](db_url())



from typing import Callable, Dict

import ckan.model as model
import ckan.plugins.toolkit as tk

import ckanext.drupal_idp.utils as utils

_actions: Dict[str, Callable] = {}

def action(func: Callable):
    _actions[f'drupal_idp_{func.__name__}'] = func
    return func


def get_actions():
    return _actions


@action
@tk.side_effect_free
def user_show(context, data_dict):
    tk.check_access('drupal_idp_user_show', context, data_dict)
    id: utils.DrupalId = tk.get_or_bust(data_dict, 'id')
    user = (
        model.Session.query(model.User.id)
        .filter(model.User.plugin_extras["drupal_idp"]["id"].astext == str(id))
        .one_or_none()
    )
    if user is None:
        raise tk.ObjectNotFound(tk._(f'DrupalId({id}) not found'))
    data_dict['id'] = user.id
    return tk.get_action("user_show")(context, data_dict)



import pytest

import ckan.model as model
import ckan.lib.munge as munge


import ckanext.drupal_idp.utils as utils


class TestDetails:
    def test_init(self, details_data):
        details = utils.Details(**details_data)
        for prop in details_data:
            assert getattr(details, prop) == details_data[prop]

    def test_into_user(self, details_data):
        details = utils.Details(**details_data)
        userdict = details.make_userdict()

        assert userdict == {
            "email": details_data["email"],
            "name": munge.munge_name(details_data["name"]),
            "sysadmin": False,
            "plugin_extras": {"drupal_idp": details_data},
        }

    def test_sysadmin_ignored_by_default(self, details_data):
        details_data["roles"].append(utils.DEFAULT_ADMIN_ROLE)
        assert not utils.Details(**details_data).is_sysadmin()

    @pytest.mark.ckan_config(utils.CONFIG_INHERIT_ADMIN_ROLE, "true")
    def test_sysadmin(self, details_data):
        assert not utils.Details(**details_data).is_sysadmin()
        details_data["roles"].append("not-an-admin")
        assert not utils.Details(**details_data).is_sysadmin()
        details_data["roles"].append(utils.DEFAULT_ADMIN_ROLE)
        assert utils.Details(**details_data).is_sysadmin()

    @pytest.mark.ckan_config(utils.CONFIG_INHERIT_ADMIN_ROLE, "true")
    @pytest.mark.ckan_config(utils.CONFIG_ADMIN_ROLE_NAME, "custom-admin")
    def test_sysadmin_with_custom_role(self, details_data):
        details_data["roles"].append(utils.DEFAULT_ADMIN_ROLE)
        assert not utils.Details(**details_data).is_sysadmin()
        details_data["roles"].append("custom-admin")
        assert utils.Details(**details_data).is_sysadmin()


class TestSynchronizationEnabled:
    def test_default_value(self):
        assert not utils.is_synchronization_enabled()

    @pytest.mark.ckan_config(utils.CONFIG_SYNCHRONIZATION_ENABLED, "true")
    def test_updated_value(self):
        assert utils.is_synchronization_enabled()


class TestSessionCookieName:
    @pytest.mark.usefixtures("with_request_context")
    def test_http(self):
        assert (
            utils.session_cookie_name()
            == "SESS49960de5880e8c687434170f6476605b"
        )

    @pytest.mark.usefixtures("with_request_context")
    def test_https(self, ckan_config, monkeypatch):
        monkeypatch.setitem(ckan_config, "ckan.site_url", "https://test.net")
        assert (
            utils.session_cookie_name()
            == "SSESS49960de5880e8c687434170f6476605b"
        )

    @pytest.mark.ckan_config(utils.CONFIG_STATIC_HOST, "my.site.com")
    def test_static_host(self):
        assert (
            utils.session_cookie_name()
            == "SESSe181b034216cdf301d365b2fc8aa54db"
        )


def test_decode_sid():
    cookie = "x7kHcLGoGrro66NTty7EPBluIVK8HNUOMvVpUT_6g7o"
    expected = "y7ZGy1KGrckATBB5gixhyKBS2BfhMeLmdMuX4_H-Uak"
    assert utils.decode_sid(cookie) == expected


class TestFixtures:
    def test_no_session(self, with_no_drupal_session):
        assert utils.get_user_details("sid") is None

    def test_with_session(self, with_drupal_session):
        assert utils.get_user_details("sid") is not None


@pytest.mark.usefixtures("clean_db")
class TestGetOrCreation:
    def test_user_created(self, details_data):
        assert model.User.get(details_data["name"]) is None
        details = utils.Details(**details_data)
        utils.get_or_create_from_details(details)
        user = model.User.get(details_data["name"])
        assert user.name == details_data["name"]
        assert user.plugin_extras["drupal_idp"] == details_data


@pytest.mark.usefixtures("clean_db")
class TestGetOrCreation:

    def test_default_native_id(self, details_data):
        details = utils.Details(**details_data)
        userdict = utils.get_or_create_from_details(details)
        assert userdict["id"] != details_data["id"]

    @pytest.mark.ckan_config(utils.CONFIG_SAME_ID, "true")
    def test_same_id(self, details_data):
        details = utils.Details(**details_data)
        userdict = utils.get_or_create_from_details(details)
        assert userdict["id"] == str(details_data["id"])

    def test_user_sync(self, details_data, monkeypatch, ckan_config):
        details = utils.Details(**details_data)
        userdict = utils.get_or_create_from_details(details)
        user = model.User.get(userdict["id"])
        user.email = "hello@world"
        user.name = "hello"
        model.Session.commit()

        userdict = utils.get_or_create_from_details(details)
        assert userdict["name"] != details_data["name"]
        assert userdict["name"] == "hello"
        user = model.User.get(userdict["id"])
        assert user.email != details_data["email"]
        assert user.email == "hello@world"

        utils.synchronize(userdict, details)
        userdict = utils.get_or_create_from_details(details)
        assert userdict["name"] == details_data["name"]
        user = model.User.get(userdict["id"])
        assert user.email == details_data["email"]

        assert (
            model.Session.query(model.User).filter_by(name="hello").count()
            == 0
        )
        assert (
            model.Session.query(model.User)
            .filter_by(name=details_data["name"])
            .count()
            == 1
        )

    def test_plugin_extras_not_erased(self, details_data):
        details = utils.Details(**details_data)
        userdict = utils.get_or_create_from_details(details)
        user = model.User.get(userdict["id"])
        user.plugin_extras = {
            "test": {"key": "value"}
        }
        utils.synchronize(userdict, details, force=True)
        user = model.User.get(userdict["id"])
        assert "drupal_idp" in user.plugin_extras
        assert "test" in user.plugin_extras



from __future__ import annotations
import dataclasses
import base64
import hashlib
import logging
import six
import secrets
from typing import Any, Dict, List, Optional, TypedDict


import ckan.model as model
import ckan.plugins.toolkit as tk
import ckan.lib.munge as munge

CONFIG_DB_URL = "ckanext.drupal_idp.db_url"
CONFIG_SYNCHRONIZATION_ENABLED = "ckanext.drupal_idp.synchronization.enabled"
CONFIG_STATIC_HOST = "ckanext.drupal_idp.host"
CONFIG_INHERIT_ADMIN_ROLE = "ckanext.drupal_idp.admin_role.inherit"
CONFIG_ADMIN_ROLE_NAME = "ckanext.drupal_idp.admin_role.name"
CONFIG_DRUPAL_VERSION = "ckanext.drupal_idp.drupal.version"
CONFIG_SAME_ID = "ckanext.drupal_idp.same_id"

DEFAULT_ADMIN_ROLE = "administrator"
DEFAULT_DRUPAL_VERSION = "9"

log = logging.getLogger(__name__)

DrupalId = int
UserDict = Dict[str, Any]


class DetailsData(TypedDict):
    name: str
    email: str
    id: DrupalId
    roles: List[str]


@dataclasses.dataclass
class Details:
    name: str
    email: str
    id: DrupalId
    roles: List[str] = dataclasses.field(default_factory=list)

    def is_sysadmin(self):
        return (
            tk.asbool(tk.config.get(CONFIG_INHERIT_ADMIN_ROLE))
            and tk.config.get(CONFIG_ADMIN_ROLE_NAME, DEFAULT_ADMIN_ROLE)
            in self.roles
        )

    def make_userdict(self):
        return {
            "email": self.email,
            "name": munge.munge_name(self.name),
            "sysadmin": self.is_sysadmin(),
            "plugin_extras": {"drupal_idp": dataclasses.asdict(self)},
        }


def is_synchronization_enabled() -> bool:
    return tk.asbool(tk.config.get(CONFIG_SYNCHRONIZATION_ENABLED))


def _make_password():
    return secrets.token_urlsafe(60)


def _get_host() -> str:
    host = tk.config.get(CONFIG_STATIC_HOST)
    if not host:
        host = tk.request.environ["HTTP_HOST"].split(":")[0]
    return host


def session_cookie_name() -> str:
    """Compute name of the cookie that stores Drupal's SessionID.

    For D9 it's PREFIX + HASH, where:
      PREFIX: if isHTTPS then SSESS else SESS
      HASH: first 32 characters of sha256 hash of the site's hostname
            (does not include port)

    """
    server_name = _get_host()
    hash = hashlib.sha256(six.ensure_binary(server_name)).hexdigest()[:32]
    name = f"SESS{hash}"
    if tk.config["ckan.site_url"].startswith("https"):
        name = "S" + name
    return name


def decode_sid(cookie_sid: str) -> str:
    """Decode Drupal's session cookie and turn it into SessionID.

    This method was written around Drupal v9.1.4 release. It's
    logic is unlikely to change for D9, but it may change in the
    future major releases, so keep an eye on it and check it first
    if you are sure that session cookie is there, but CKAN can't
    obtain user from Drupal's database.

    Algorythm:
    - get cookie value
    - sha256 it
    - base64 it
    - replace pluses and slashes
    - strip out `=`-paddings

    """
    sha_hash = hashlib.sha256(six.ensure_binary(cookie_sid)).digest()
    base64_hash = base64.encodebytes(sha_hash)
    trans_rules = str.maketrans(
        {
            "+": "-",
            "/": "_",
            "=": "",
        }
    )
    sid = six.ensure_str(base64_hash.strip()).translate(trans_rules)
    return sid


def get_user_details(sid: str) -> Optional[Details]:
    """Fetch user data from Drupal's database."""
    import ckanext.drupal_idp.drupal as drupal

    adapter = drupal.get_adapter(
        tk.config.get(CONFIG_DRUPAL_VERSION, DEFAULT_DRUPAL_VERSION)
    )
    user = adapter.get_user_by_sid(sid)
    # check if session has username,
    # otherwise is unauthenticated user session
    if not user:
        return
    details_data = DetailsData(**user)
    roles = adapter.get_user_roles(user.id)
    details_data["roles"] = roles
    return Details(**details_data)



def _get_by_email(email: str) -> Optional[UserDict]:
    user = (
        model.Session.query(model.User.id)
        .filter(model.User.email == email)
        .one_or_none()
    )
    if user is None:
        return
    return tk.get_action("user_show")({"ignore_auth": True}, {"id": user.id})


def _create_from_details(details: Details) -> UserDict:
    """Create a user with random password using Drupal's data.

    Raises:
    ValidationError if email or username is not unique
    """
    user = details.make_userdict()
    user["password"] = _make_password()
    if tk.asbool(tk.config.get(CONFIG_SAME_ID)):
        user["id"] = details.id

    admin = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    user = tk.get_action("user_create")({"user": admin["name"]}, user)
    return user


def _attach_details(id: str, details: Details) -> UserDict:
    """Update name, email and plugin_extras

    Raises:
    ValidationError if email or username is not unique
    """
    admin = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    user = tk.get_action("user_show")({"user": admin["name"]}, {"id": id, "include_plugin_extras": True})

    # do not drop extras that were set by other plugins
    extras = user.pop('plugin_extras', {})
    patch = details.make_userdict()
    extras.update(patch['plugin_extras'])
    patch['plugin_extras'] = extras
    user.update(patch)

    # user_patch is not available in v2.9
    user = tk.get_action("user_update")({"user": admin["name"]}, user)
    return user


def get_or_create_from_details(details: Details) -> UserDict:
    """Get existing user or create new one.

    Raises:
    ValidationError if email or username is not unique
    """
    try:
        user = tk.get_action("drupal_idp_user_show")({"ignore_auth": True}, {"id": details.id})
    except tk.ObjectNotFound:
        user = _get_by_email(details.email)
        if user:
            user = _attach_details(user["id"], details)
    return user or _create_from_details(details)


def synchronize(user: UserDict, details: Details, force: bool = False) -> UserDict:
    userobj = model.User.get(user["id"])
    if userobj.name != details.name:
        log.info(f"Synchronizing user {userobj.name} -> {details.name}")
        if model.User.get(details.name) is not None:
            raise tk.ValidationError(
                {"name": "This username is already taken"}
            )
        userobj.name = details.name
        model.Session.commit()
    if (
        force or
        userobj.email != details.email
        or userobj.sysadmin != details.is_sysadmin()
    ):
        log.info(f"Synchronizing user {details.name}")
        user = _attach_details(user["id"], details)
    return user



import ckan.plugins.toolkit as tk

_auth_functions = {}

def auth_function(func):
    _auth_functions[f'drupal_idp_{func.__name__}'] = func
    return func


def get_auth_functions():
    return _auth_functions


@auth_function
@tk.auth_disallow_anonymous_access
def user_show(context, data_dict):
    return {'success': False, 'msg': tk._('Users are not allowed to use DrupalID')}
